!> @file user_actions.f90
!------------------------------------------------------------------------------!
! This file is part of the PALM model system.
!
! PALM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2018 Leibniz Universitaet Hannover
!------------------------------------------------------------------------------!
!
! Current revisions:
! ------------------
! 
! 
! Former revisions:
! -----------------
! $Id: user_actions.f90 2718 2018-01-02 08:49:38Z maronga $
! Corrected "Former revisions" section
! 
! 2696 2017-12-14 17:12:51Z kanani
! Change in file header (GPL part)
!
! 2101 2017-01-05 16:42:31Z suehring
!
! 2000 2016-08-20 18:09:15Z knoop
! Forced header and separation lines into 80 columns
! 
! 1960 2016-07-12 16:34:24Z suehring
! New CASE statement for scalar tendency
! 
! 1873 2016-04-18 14:50:06Z maronga
! Module renamed (removed _mod)
! 
! 
! 1850 2016-04-08 13:29:27Z maronga
! Module renamed
!
! 
! 1822 2016-04-07 07:49:42Z hoffmann
! qr/nr-tendency removed.
!
! 1682 2015-10-07 23:56:08Z knoop
! Code annotations made doxygen readable 
! 
! 1353 2014-04-08 15:21:23Z heinze
! REAL constants provided with KIND-attribute
! 
! 1320 2014-03-20 08:40:49Z raasch
! kind-parameters added to all INTEGER and REAL declaration statements,
! kinds are defined in new module kinds,
! revision history before 2012 removed,
! comment fields (!:) to be used for variable explanations added to
! all variable declaration statements 
!
! 1318 2014-03-17 13:35:16Z raasch
! module interfaces removed
!
! 1053 2012-11-13 17:11:03Z hoffmann
! +qr-tendency, nr-tendency
!
! 1036 2012-10-22 13:43:42Z raasch
! code put under GPL (PALM 3.9)
!
! 211 2008-11-11 04:46:24Z raasch
! Former file user_interface.f90 split into one file per subroutine
!
! Description:
! ------------
!> Execution of user-defined actions before or after single timesteps
!------------------------------------------------------------------------------!
 MODULE user_actions_mod
 

    PRIVATE
    PUBLIC user_actions

    INTERFACE user_actions
       MODULE PROCEDURE user_actions
       MODULE PROCEDURE user_actions_ij
    END INTERFACE user_actions

 CONTAINS


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Call for all grid points
!------------------------------------------------------------------------------!
    SUBROUTINE user_actions( location )

       USE control_parameters

       USE cpulog

       USE indices

       USE kinds

       USE pegrid

       USE user

       USE arrays_3d

       USE basic_constants_and_equations_mod,                                  &
           ONLY:  pi

       IMPLICIT NONE

       CHARACTER (LEN=*) ::  location !< 

       INTEGER(iwp) ::  i !< 
       INTEGER(iwp) ::  j !< 
       INTEGER(iwp) ::  k !< 
       INTEGER(iwp) ::  n !< 

       INTEGER(iwp) ::  ii
       INTEGER(iwp) ::  idx
       REAL(wp)     ::  theta ! constant change in WD in deg/h
       REAL(wp)     ::  F_tmp ! wind speed
       REAL(wp)     ::  D_tmp ! wind direction
   
       REAL(wp) :: pt_new = 9999999.9_wp 

       INTEGER(iwp) ::  ngp, sender, ind(4)
       
       REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  local_2d_u, local_2d_v, local_2d_w,  &
                                             local_2d_e, local_2d_pt
					     
       REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  local_pf	
       
       REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  total_2d_u, total_2d_v, total_2d_w,  &
                                             total_2d_e, total_2d_pt
       
       REAL(wp) ::  output_layer	

       CALL cpu_log( log_point(24), 'user_actions', 'start' )

!
!--    Here the user-defined actions follow
!--    No calls for single grid points are allowed at locations before and
!--    after the timestep, since these calls are not within an i,j-loop
       SELECT CASE ( location )

          CASE ( 'before_timestep' )
!
!--          Enter actions to be done before every timestep here
!



!--          Wind direction change of geostrophic wind
!--          Determine theta (WD change gradient) for current time step
             IF ( theta_0 == 999. ) THEN
                idx = MINLOC( ABS( WD_time - simulated_time ) ,DIM=1)
                theta = WD_theta(idx)
             ELSE
                theta = theta_0
             ENDIF

!--          Adjust Geostrophic wind
             DO ii = 0, SIZE(ug)-1 
                F_tmp = (ug(ii)**2_wp+vg(ii)**2_wp)**0.5_wp
                D_tmp = ATAN(vg(ii)/ug(ii))
                D_tmp = D_tmp + (theta * pi / 180_wp / 3600_wp * -1_wp ) * dt_3d
                ug(ii) = COS(D_tmp) * F_tmp
                vg(ii) = SIN(D_tmp) * F_tmp
             ENDDO

          CASE ( 'after_integration' )
!
!--          Enter actions to be done after every time integration (before
!--          data output)
           IF ( simulated_time >= 0.0 ) THEN

             IF ( .NOT. ALLOCATED( local_2d_u ) ) THEN	  
	        ALLOCATE( local_2d_u(nysg:nyng,nzb:nzt+1) )
   	     END IF
	     IF ( .NOT. ALLOCATED( local_2d_v ) ) THEN
	        ALLOCATE( local_2d_v(nysg:nyng,nzb:nzt+1) )
	     END IF
	     IF ( .NOT. ALLOCATED( local_2d_w ) ) THEN 		
	        ALLOCATE( local_2d_w(nysg:nyng,nzb:nzt+1) )
	     END IF
	     IF ( .NOT. ALLOCATED( local_2d_e ) ) THEN	
 	        ALLOCATE( local_2d_e(nysg:nyng,nzb:nzt+1) )
	     END IF
	     IF ( .NOT. ALLOCATED( local_2d_pt ) ) THEN	
	        ALLOCATE( local_2d_pt(nysg:nyng,nzb:nzt+1) )
	     END IF	

             IF ( myid == 0 ) THEN
	        IF ( .NOT. ALLOCATED( total_2d_u ) ) THEN	  
	           ALLOCATE( total_2d_u(-nbgp:ny+nbgp,nzb:nzt+1) )
	        END IF
	        IF ( .NOT. ALLOCATED( total_2d_v ) ) THEN	
	           ALLOCATE( total_2d_v(-nbgp:ny+nbgp,nzb:nzt+1) )
 	        END IF
	        IF ( .NOT. ALLOCATED( total_2d_w ) ) THEN	
	           ALLOCATE( total_2d_w(-nbgp:ny+nbgp,nzb:nzt+1) )
	        END IF
	        IF ( .NOT. ALLOCATED( total_2d_e ) ) THEN	
	           ALLOCATE( total_2d_e(-nbgp:ny+nbgp,nzb:nzt+1) )
	        END IF
	        IF ( .NOT. ALLOCATED( total_2d_pt ) ) THEN	
	           ALLOCATE( total_2d_pt(-nbgp:ny+nbgp,nzb:nzt+1) )
	        END IF
	     END IF

!--          Implementation einer Ausgabe von yz-Schnitten der Geschwindigkeitskomponenten 
!--          u, v und w, der potentiellen Temperatur pt und der subskaligen turbulenten 
!--          kinetischen Energie e  
!--          Die Vorgehensweise folgt hier der Vorgehensweise bei der Ausgabe von 
!--          yz-Schnitten in data_output_2d.f90

             output_layer = nx - 1

!--          Allocate a temporary array for resorting (kji -> ijk)
             IF ( .NOT. ALLOCATED( local_pf ) ) THEN
                ALLOCATE( local_pf(nxlg:nxrg,nysg:nyng,nzb:nzt+1) )
             END IF		

!--          Output erfolgt nur, wenn yz-Ebene zumindest teilweise auf Kern vorliegt
             IF ( ( output_layer >= nxl ) .AND. ( output_layer <= nxr ) ) THEN
!--             Eventuell ist an dieser Stelle noch ein Umspeichern von Feldern 
!--             notwendig, damit später auszutauschende Informationen  zusammenhaengend
!--           	vorliegen
!--             Bei den folgenden Feldern muss die Reihenfolge der Indizes i,j,k lauten 
!--             und damit eben genau umgekehrt sein, wie in den normalerweise 
!--             vorliegenden Feldern 

                DO  i = nxlg, nxrg
                   DO  j = nysg, nyng
                      DO  k = nzb, nzt+1
                         local_pf(i,j,k) = u(k,j,i)
                      ENDDO
                   ENDDO
                ENDDO
	        local_2d_u  = local_pf(output_layer,:,nzb:nzt+1)
		
		DO  i = nxlg, nxrg
		   DO  j = nysg, nyng
	 	      DO  k = nzb, nzt+1
		         local_pf(i,j,k) = v(k,j,i)
		      ENDDO
		   ENDDO
		ENDDO        
		local_2d_v  = local_pf(output_layer,:,nzb:nzt+1)
		
		DO  i = nxlg, nxrg
		   DO  j = nysg, nyng
		      DO  k = nzb, nzt+1
		         local_pf(i,j,k) = w(k,j,i)
		      ENDDO
		   ENDDO
		ENDDO       
		local_2d_w  = local_pf(output_layer,:,nzb:nzt+1)
		
		DO  i = nxlg, nxrg
		   DO  j = nysg, nyng
		      DO  k = nzb, nzt+1
		         local_pf(i,j,k) = e(k,j,i)
		      ENDDO
		   ENDDO
		ENDDO       
		local_2d_e  = local_pf(output_layer,:,nzb:nzt+1)
		
		DO  i = nxlg, nxrg
		   DO  j = nysg, nyng
		      DO  k = nzb, nzt+1
		         local_pf(i,j,k) = pt(k,j,i)
	              ENDDO
		   ENDDO
		ENDDO      		 
		local_2d_pt = local_pf(output_layer,:,nzb:nzt+1)
	     END IF
	     
!
!--          PE0 receives partial arrays from all processors of the 
!--          respective cross section and outputs them. Here a
!--          barrier has to be set, because otherwise 
!--          "-MPI- FATAL: Remote protocol queue full" may occur.
             CALL MPI_BARRIER( comm2d, ierr )

             ngp = ( nyng-nysg+1 ) * ( nzt-nzb+2 )
             IF ( myid == 0 )  THEN
!
!--             Local array can be relocated directly.
                IF ( ( output_layer >= nxl   .AND.              &
                       output_layer <= nxr ) ) THEN
                   total_2d_u(nysg:nyng,nzb:nzt+1)  = local_2d_u
		   total_2d_v(nysg:nyng,nzb:nzt+1)  = local_2d_v
		   total_2d_w(nysg:nyng,nzb:nzt+1)  = local_2d_w
		   total_2d_e(nysg:nyng,nzb:nzt+1)  = local_2d_e
		   total_2d_pt(nysg:nyng,nzb:nzt+1) = local_2d_pt
                END IF
		
!
!--             Receive data from all other PEs.
                DO  n = 1, numprocs-1
!
!--                Receive index limits first, then array.
!--                Index limits are received in arbitrary order from
!--                the PEs.
                   CALL MPI_RECV( ind(1), 4, MPI_INTEGER,     &
                                  MPI_ANY_SOURCE, 0, comm2d,  &
                                  status, ierr )
	  
!
!--                Not all PEs have data for YZ-cross-section.
                   IF ( ind(1) /= -9999 )  THEN
                      sender = status(MPI_SOURCE)
		      IF ( ALLOCATED( local_2d_u ) ) THEN
                         DEALLOCATE( local_2d_u )
		      END IF
		      IF ( ALLOCATED( local_2d_v ) ) THEN	 
		         DEALLOCATE( local_2d_v )
		      END IF
		      IF ( ALLOCATED( local_2d_w ) ) THEN	 
		         DEALLOCATE( local_2d_w )
		      END IF
		      IF ( ALLOCATED( local_2d_e ) ) THEN	 
		         DEALLOCATE( local_2d_e )
		      END IF
		      IF ( ALLOCATED( local_2d_pt ) ) THEN	 
		         DEALLOCATE( local_2d_pt )
		      END IF
	 
                      ALLOCATE( local_2d_u(ind(1):ind(2), &
                                ind(3):ind(4)) )
	              ALLOCATE( local_2d_v(ind(1):ind(2), &
				ind(3):ind(4)) )
		      ALLOCATE( local_2d_w(ind(1):ind(2), &
				ind(3):ind(4)) )
		      ALLOCATE( local_2d_e(ind(1):ind(2), &
				ind(3):ind(4)) )
		      ALLOCATE( local_2d_pt(ind(1):ind(2), &
				ind(3):ind(4)) )
		       		       		       	       
                      CALL MPI_RECV( local_2d_u(ind(1),ind(3)), ngp,  &
                                     MPI_REAL, sender, 1, comm2d,     &
                                     status, ierr )
	     
		      CALL MPI_RECV( local_2d_v(ind(1),ind(3)), ngp,  &
				     MPI_REAL, sender, 2, comm2d,     &
				     status, ierr )
	     
		      CALL MPI_RECV( local_2d_w(ind(1),ind(3)), ngp,  &
				     MPI_REAL, sender, 3, comm2d,     &
				     status, ierr )
	     
		      CALL MPI_RECV( local_2d_e(ind(1),ind(3)), ngp,  &
				     MPI_REAL, sender, 4, comm2d,     &
				     status, ierr )
		     
		      CALL MPI_RECV( local_2d_pt(ind(1),ind(3)), ngp, &
				     MPI_REAL, sender, 5, comm2d,     &
				     status, ierr )
	     		 		   		 		 
                      total_2d_u(ind(1):ind(2),ind(3):ind(4)) =              &
                                                                 local_2d_u
		      total_2d_v(ind(1):ind(2),ind(3):ind(4)) =              &
				                                 local_2d_v
		      total_2d_w(ind(1):ind(2),ind(3):ind(4)) =              &
				                                 local_2d_w
		      total_2d_e(ind(1):ind(2),ind(3):ind(4)) =              &
				                                 local_2d_e
		      total_2d_pt(ind(1):ind(2),ind(3):ind(4)) =             &
				                     	         local_2d_pt
					 		    					    					    					    
                   ENDIF
                ENDDO
	
!
!--             Output of the total cross-section.
!                WRITE (819)  total_2d_u(0:ny+1,nzb:nzt+1)
                WRITE (819)  total_2d_u(-nbgp:ny+nbgp,nzb:nzt+1)
		PRINT*,simulated_time, total_2d_u(10,10)
		
!		 WRITE (820)  total_2d_v(0:ny+1,nzb:nzt+1)
                WRITE (820)  total_2d_v(-nbgp:ny+nbgp,nzb:nzt+1)
		
! 		 WRITE (821)  total_2d_w(0:ny+1,nzb:nzt+1)
                WRITE (821)  total_2d_w(-nbgp:ny+nbgp,nzb:nzt+1)
		
!		 WRITE (822)  total_2d_e(0:ny+1,nzb:nzt+1)
                WRITE (822)  total_2d_e(-nbgp:ny+nbgp,nzb:nzt+1)
		
!	 	 WRITE (823)  total_2d_pt(0:ny+1,nzb:nzt+1)
                WRITE (823)  total_2d_pt(-nbgp:ny+nbgp,nzb:nzt+1)
!
!--             Relocate the local array for the next loop increment
                IF ( ALLOCATED( local_2d_u ) ) THEN
                   DEALLOCATE( local_2d_u )
		END IF
		IF ( ALLOCATED( local_2d_v ) ) THEN  
		   DEALLOCATE( local_2d_v )
		END IF
		IF ( ALLOCATED( local_2d_w ) ) THEN   
		   DEALLOCATE( local_2d_w )
		END IF
		IF ( ALLOCATED( local_2d_e ) ) THEN  
		   DEALLOCATE( local_2d_e )
		END IF   
		IF ( ALLOCATED( local_2d_pt ) ) THEN   
		   DEALLOCATE( local_2d_pt )
		END IF   
		
		IF ( ALLOCATED( total_2d_u ) ) THEN
                   DEALLOCATE( total_2d_u )
		END IF
		IF ( ALLOCATED( total_2d_v ) ) THEN  
		   DEALLOCATE( total_2d_v )
		END IF
		IF ( ALLOCATED( total_2d_w ) ) THEN  
		   DEALLOCATE( total_2d_w )
		END IF
		IF ( ALLOCATED( total_2d_e ) ) THEN   
		   DEALLOCATE( total_2d_e )
		END IF
		IF ( ALLOCATED( total_2d_pt ) ) THEN   
		   DEALLOCATE( total_2d_pt )
		END IF   		

             ELSE
!
!--             If the cross section resides on the PE, send the
!--             local index limits, otherwise send -9999 to PE0.
                IF ( ( output_layer >= nxl  .AND. output_layer <= nxr ) ) THEN
                   ind(1) = nysg;  ind(2) = nyng
                   ind(3) = nzb;   ind(4) = nzt+1
                ELSE
                   ind(1) = -9999; ind(2) = -9999
                   ind(3) = -9999; ind(4) = -9999
                ENDIF
                CALL MPI_SEND( ind(1), 4, MPI_INTEGER, 0, 0, &
                               comm2d, ierr )
			       
!
!--             If applicable, send data to PE0.
                IF ( ind(1) /= -9999 )  THEN
		   IF ( .NOT. ALLOCATED( local_2d_u ) ) THEN
		      ALLOCATE( local_2d_u(ind(1):ind(2), &
                                ind(3):ind(4)) )
		   END IF
		   IF ( .NOT. ALLOCATED( local_2d_v ) ) THEN
		      ALLOCATE( local_2d_v(ind(1):ind(2), &
		                ind(3):ind(4)) )
	           END IF
		   IF ( .NOT. ALLOCATED( local_2d_w ) ) THEN
		      ALLOCATE( local_2d_w(ind(1):ind(2), &
		                ind(3):ind(4)) )
		   END IF
		   IF ( .NOT. ALLOCATED( local_2d_e ) ) THEN
		      ALLOCATE( local_2d_e(ind(1):ind(2), &
		                ind(3):ind(4)) )
		   END IF
		   IF ( .NOT. ALLOCATED( local_2d_pt ) ) THEN
		      ALLOCATE( local_2d_pt(ind(1):ind(2), &
		                ind(3):ind(4)) )
		   END IF										
		
                   CALL MPI_SEND( local_2d_u(nysg,nzb), ngp, &
                                  MPI_REAL, 0, 1, comm2d, ierr )
                   CALL MPI_SEND( local_2d_v(nysg,nzb), ngp, &
                                  MPI_REAL, 0, 2, comm2d, ierr )
                   CALL MPI_SEND( local_2d_w(nysg,nzb), ngp, &
                                  MPI_REAL, 0, 3, comm2d, ierr )
                   CALL MPI_SEND( local_2d_e(nysg,nzb), ngp, &
                                  MPI_REAL, 0, 4, comm2d, ierr )
                   CALL MPI_SEND( local_2d_pt(nysg,nzb), ngp, & 
                                  MPI_REAL, 0, 5, comm2d, ierr )
				  
		   IF ( ALLOCATED(local_2d_u) ) THEN
		      DEALLOCATE( local_2d_u )		      
		   END IF
		   IF ( ALLOCATED(local_2d_v) ) THEN
		      DEALLOCATE( local_2d_v )
		   END IF
		   IF ( ALLOCATED(local_2d_w) ) THEN
		      DEALLOCATE( local_2d_w )
		   END IF
		   IF ( ALLOCATED(local_2d_e) ) THEN
		      DEALLOCATE( local_2d_e )
		   END IF
		   IF ( ALLOCATED(local_2d_pt) ) THEN
		      DEALLOCATE( local_2d_pt )
		   END IF               		  
				  
                ENDIF
             ENDIF
	     
!
!--          A barrier has to be set, because otherwise some PEs may
!--          proceed too fast so that PE0 may receive wrong data on
!--          tag 0
             CALL MPI_BARRIER( comm2d, ierr )	
	     
	     IF ( ALLOCATED( local_pf ) ) THEN
	        DEALLOCATE( local_pf )
	     END IF	
	     
	     CALL MPI_BARRIER( comm2d, ierr )
	     
	     CALL MPI_BARRIER( comm2d, ierr )
	     
	   END IF

          CASE ( 'after_timestep' )
!
!--          Enter actions to be done after every timestep here


          CASE ( 'u-tendency' )
!
!--          Enter actions to be done in the u-tendency term here


          CASE ( 'v-tendency' )


          CASE ( 'w-tendency' )


          CASE ( 'pt-tendency' )


          CASE ( 'sa-tendency' )


          CASE ( 'e-tendency' )


          CASE ( 'q-tendency' )
          
          
          CASE ( 's-tendency' )          


          CASE DEFAULT
             message_string = 'unknown location "' // location // '"'
             CALL message( 'user_actions', 'UI0001', 1, 2, 0, 6, 0 )

       END SELECT

       CALL cpu_log( log_point(24), 'user_actions', 'stop' )

    END SUBROUTINE user_actions


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Call for grid point i,j
!------------------------------------------------------------------------------!
    SUBROUTINE user_actions_ij( i, j, location )

       USE control_parameters
       USE kinds
       USE pegrid
       USE user
       USE indices, &
          ONLY: nzb, nzt
       USE arrays_3d, &
          ONLY: tend,u,v
       USE basic_constants_and_equations_mod,                                     &
           ONLY:  pi
       IMPLICIT NONE

       CHARACTER (LEN=*) ::  location
       INTEGER(iwp) ::  idum
       INTEGER(iwp) ::  k,j,i
       INTEGER(iwp) ::  idx
       REAL(wp)     ::  theta ! constant change in WD in deg/h

!
!--    Here the user-defined actions follow
       SELECT CASE ( location )

          CASE ( 'u-tendency' )
!--           Wind direction change on u-component
              IF ( theta_0 == 999. ) THEN
                  idx = MINLOC( ABS( WD_time - simulated_time ) ,DIM=1)
                  theta = WD_theta(idx)
              ELSE
                 theta = theta_0
              ENDIF
              
              DO k = nzb+1,nzt 
                 tend(k,j,i) = tend(k,j,i) - theta * -1_wp * pi / 180_wp / 3600_wp * v(k,j,i)
              ENDDO
              

          CASE ( 'v-tendency' )
!--           Wind direction change on v-component
              IF ( theta_0 == 999. ) THEN
                  idx = MINLOC( ABS( WD_time - simulated_time ) ,DIM=1)
                  theta = WD_theta(idx)
              ELSE
                  theta = theta_0
              ENDIF
              
              DO k = nzb+1,nzt   
                 tend(k,j,i) = tend(k,j,i) + theta * -1_wp * pi / 180_wp / 3600_wp * u(k,j,i)
              ENDDO

          CASE ( 'w-tendency' )


          CASE ( 'pt-tendency' )


          CASE ( 'sa-tendency' )


          CASE ( 'e-tendency' )


          CASE ( 'q-tendency' )
          
          
          CASE ( 's-tendency' )


          CASE ( 'before_timestep', 'after_integration', 'after_timestep' )
             message_string = 'location "' // location // '" is not ' // &
                             'allowed to be called with parameters "i" and "j"'
             CALL message( 'user_actions', 'UI0002', 1, 2, 0, 6, 0 )


          CASE DEFAULT
             message_string = 'unknown location "' // location // '"'
             CALL message( 'user_actions', 'UI0001', 1, 2, 0, 6, 0 )
             

       END SELECT

    END SUBROUTINE user_actions_ij

 END MODULE user_actions_mod
